/******************************************************************************
**
**  Project Name:	DropShell
**     File Name:	DSUserProcs.c
**
**   Description:	Specific AppleEvent handlers used by the DropBox
**
*******************************************************************************
**                       A U T H O R   I D E N T I T Y
*******************************************************************************
**
**	Initials	Name
**	--------	-----------------------------------------------
**	LDR			Leonard Rosenthol
**	MTC			Marshall Clow
**	SCS			Stephan Somogyi
**
*******************************************************************************
**                      R E V I S I O N   H I S T O R Y
*******************************************************************************
**
**	  Date		Time	Author	Description
**	--------	-----	------	---------------------------------------------
**	06/23/94			LDR		Added support for ProcessItem and ProcessFolder handling
**	02/20/94			LDR		Modified Preflight & Postflight to take item count
**	01/25/92			LDR		Removed the use of const on the userDataHandle
**	12/09/91			LDR		Added the new SelectFile userProc
**								Added the new Install & DisposeUserGlobals procs
**								Modified PostFlight to only autoquit on odoc, not pdoc
**	11/24/91			LDR		Added the userProcs for pdoc handler
**								Cleaned up the placement of braces
**								Added the passing of a userDataHandle
**	10/29/91			SCS		Changes for THINK C 5
**	10/28/91			LDR		Officially renamed DropShell (from QuickShell)
**								Added a bunch of comments for clarification
**	10/06/91	00:02	MTC		Converted to MPW C
**	04/09/91	00:02	LDR		Added to Projector
**
******************************************************************************/

#include <StandardFile.h>
#include <FixMath.h>
#include <Timer.h>			// For timing

#ifndef TARGET_API_MAC_CARBON
#include <ICTypes.h>
#include <ICKeys.h>
#include <ICAPI.h>
#else
typedef long ICError;										/* type for error codes */
#endif

#include "DSGlobals.h"
#include "DSUserProcs.h"

#include "applefile.h"

#include <StuTypes.h>

// Static Prototypes
static OSErr ProcessItem(FSSpecPtr myFSSPtr);
static OSErr ProcessFolder(FSSpecPtr myFSSPtr);

/*
	Uncomment this line if you want each item of a dropped folder processed
	as an individual item
*/
#define qWalkFolders


/*
	This routine is called during init time.
	
	It allows you to install more AEVT Handlers beyond the standard four
*/
#pragma segment Main
pascal void InstallOtherEvents(void) {
}


/*	
	This routine is called when an OAPP event is received.
	
	Currently, all it does is set the gOApped flag, so you know that
	you were called initally with no docs, and therefore you shouldn't
	quit when done processing any following odocs.
*/
#pragma segment Main
pascal void OpenApp(void) {
	gOApped = true;
}


/*	
	This routine is called when an QUIT event is received.
	
	We simply set the global done flag so that the main event loop can
	gracefully exit.  We DO NOT call ExitToShell for two reasons:
	1) It is a pretty ugly thing to do, but more importantly
	2) The Apple event manager will get REAL upset!
*/
#pragma segment Main
pascal void QuitApp(void) {
	gDone = true;	/*	All Done! */
}


#define DO_TIMING 0

#define IsError(X) ((X) !=1 && (X) != 0)
#define Alert_NoMemory  130
#define Alert_NoIC      131
#define Alert_DiskError 132

#if DO_TIMING
static UnsignedWide read_start_time, read_end_time, read_time;
static UnsignedWide create_start_time, create_end_time, create_time;
static UnsignedWide write_start_time, write_end_time, write_time;
static UnsignedWide close_start_time, close_end_time, close_time;
static UnsignedWide total_start_time, total_end_time;
#endif

typedef struct
	{
	HParamBlockRec io;
	FSSpec *theSpec;
	unsigned char output_namesuffix;
	unsigned char output_namesuffix_length;
	volatile Boolean FileBeingCreated;
	void *GlobalsReg;
	} Create_HParamBlockRec;

static Handle InBuffer;					// Handle to input buffer
static Handle OutBuffer1, OutBuffer2;	// Handles to output output buffers
static Handle OutBuffer;				// Handle to current output buffer
static long InHandleSize;				// Size of input buffer
static long OutHandleSize;				// Size of output buffer
static short ReadRefNum;				// The file being read
static long BytesRemaining;				// Bytes of file still unread
static Ptr InPtr = NULL;				// Next character to read from buffer
static Ptr InEnd = NULL;				// End of data in buffer
static char *CurrentLine;				// Pointer to current line in the inbuffer
static int LineLength;					// Length of current line
static char TempLineBuffer[256];		// Temp holding buffer for lines spanning buffers
static Ptr OutPtr;						// Next byte to written to buffer
static Ptr OutEnd;						// End output buffer

static WindowPtr StatusWindow;
static ICInstance IC_instance = NULL;
static Handle     IC_mappings = NULL;

static HParamBlockRec Async_Info_PB, Async_Open_PB;
static ParamBlockRec Async_Write_PB, Async_Close_PB;
static FSSpec createSpec, writeSpec, closeSpec;

static unsigned char B64decode_table[256];
// B64_filename is the FULL name contained on the last "Content-Type... name= ..." line
// If the last thing we created was a resource fork then B64_lastresourcefilename holds its
// FULL name (as for B64_filename) for the purpose of deciding whether to overlay a data fork.
static char          B64_filename[256];					// C string
static char          B64_lastresourcefilename[256];		// C string
static unsigned char B64_MIMEtype[256];					// Pascal string
static OSType        MacTypeCode;
static OSType        CreatorCode;
Boolean              NeedAppleSingleHeader;
Boolean              SetFileInfoAfterClose;

static const Point zeroPoint;

static void MakeStatusWindow(void)
	{
	Rect bounds;
#ifndef TARGET_API_MAC_CARBON
	short screencentre = (qd.screenBits.bounds.right - qd.screenBits.bounds.left) / 2;
#else
	short screencentre = 320;
#endif
	bounds.top    = 100;
	bounds.bottom = 290;
	bounds.left   = screencentre - 200;
	bounds.right  = screencentre + 200;
	StatusWindow = NewWindow(NULL, &bounds, NULL, FALSE, dBoxProc, (WindowPtr)-1, FALSE, 0);
	}

static void PaintStatusWindow(void)
	{
	SetPort(GetWindowPort(StatusWindow));
	TextSize(12);
	TextFace(bold);
	MoveTo(20, 30); DrawString("\pFast, Simple, Stupid UUDecoder");
	MoveTo(20, 50); DrawString("\pby Stuart Cheshire <cheshire@cs.stanford.edu>");
	MoveTo(20, 70); DrawString("\pThis software is free.");
	MoveTo(20, 90); DrawString("\pYou can do whatever you like with it.");
	TextFace(0);
	MoveTo(20,110); DrawString("\pBuilt using DropShell 2.0");
	MoveTo(20,130); DrawString("\pby Leonard Rosenthol, Marshall Clow, and Stephan Somogyi");
	}

static void UpdateStatusWindowInput(unsigned char *inputname)
	{
	Rect r;// = StatusWindow->portRect;
	GetPortBounds(GetWindowPort(StatusWindow), &r);
	r.top    = 135;
	r.bottom = 155;
	r.left   = 20;
	EraseRect(&r);
	TextFace(bold);
	MoveTo(20,150); DrawString("\pReading File: "); DrawString(inputname);
	}

static void UpdateStatusWindowOutput(unsigned char *outputname)
	{
	Rect r;// = StatusWindow->portRect;
	GetPortBounds(GetWindowPort(StatusWindow), &r);
	r.top    = 155;
	r.left   = 20;
	EraseRect(&r);
	TextFace(bold);
	MoveTo(20,170); DrawString("\pWriting File: "); DrawString(outputname);
	}

static void UpdateStatusWindow(Boolean reading)
	{
	Rect r;// = StatusWindow->portRect;
	GetPortBounds(GetWindowPort(StatusWindow), &r);
	r.top   = 135;
	r.right =  20;
	EraseRect(&r);
	if (reading) MoveTo(8,150); else MoveTo(8,170);
	TextFace(0);
	DrawString("\p");
	}

static void StUU_Exit(void)
	{
	if (InBuffer )    { DisposeHandle(InBuffer );    InBuffer      = NULL; }
	if (OutBuffer1)   { DisposeHandle(OutBuffer1);   OutBuffer1    = NULL; }
	if (OutBuffer2)   { DisposeHandle(OutBuffer2);   OutBuffer2    = NULL; }
	if (StatusWindow) { DisposeWindow(StatusWindow); StatusWindow  = NULL; }
	if (IC_instance)  { ICStop(IC_instance);         IC_instance   = NULL; }
	}

#ifndef TARGET_API_MAC_CARBON
static Boolean TrapAvailable(unsigned short trap)
	{
	TrapType tType = (trap & 0x800 ? ToolTrap : OSTrap);
	if (trap & 0x800)				// if it is a ToolBox Trap
		{
		unsigned long n = 0x400;	// number of toolbox traps
		if (NGetTrapAddress(_InitGraf, ToolTrap) == NGetTrapAddress(0xAA6E, ToolTrap))
			n = 0x200;
		if ((trap &= 0x7FF) >= n) trap = _Unimplemented;
		}
	return(NGetTrapAddress(trap, tType) != NGetTrapAddress(_Unimplemented, ToolTrap));
	}
#else
#define TrapAvailable(X) 1
#endif

#if __MC68K__
#pragma parameter __D0 GetHoldableBytes()	// The definition of GetHoldableBytes for 68K code
extern pascal Size GetHoldableBytes(void) TWOWORDINLINE(0x70FD, 0xA05C);
#else
#ifndef TARGET_API_MAC_CARBON
static pascal Size GetHoldableBytes(void)	// The definition of GetHoldableBytes for PPC code
	{
	static unsigned short GetHoldableBytes68K[] =
		{
		0x70FD,		// moveq #-3,d0
		0xA05C,		// MemoryDispatch
		0x4E75,		// rts
		};
	enum
		{
		uppGetHoldableBytesProcInfo = kRegisterBased
		| RESULT_SIZE(kFourByteCode) | REGISTER_RESULT_LOCATION(kRegisterD0)
		};
	return(CallUniversalProc((UniversalProcPtr)GetHoldableBytes68K, uppGetHoldableBytesProcInfo));
	}
#endif
#endif

static Size sensibleTempMaxMem(Size *grow)
	{
	Size freemem = TempMaxMem(grow);
	long VMAttr, PhysicalRAM, LogicalRAM;
	
	// if VM on, see if we can revise our estimate of free memory to a more sensible value
	if (TrapAvailable(_Gestalt) &&
		Gestalt(gestaltVMAttr, &VMAttr) == noErr && (VMAttr & (1 << gestaltVMPresent)) &&
		Gestalt(gestaltPhysicalRAMSize, &PhysicalRAM) == noErr &&
		Gestalt(gestaltLogicalRAMSize, &LogicalRAM) == noErr)
		{
		// See how much physical RAM is available
		// If MemoryDispatch is available, use it, else just use
		// a simplistic estimate that half of the physical RAM is holdable
		Size holdable =
#ifndef TARGET_API_MAC_CARBON
			TrapAvailable(_MemoryDispatch) ? GetHoldableBytes() :
#endif
			PhysicalRAM / 2;

		// See how much memory is in use
		Size usedmem = LogicalRAM - TempFreeMem();

		// See how much physical RAM is free
		// If more holdable pages than currently in use, then we definitely have some usable memory
		Size useable = 0;
		if (holdable > usedmem) useable = holdable - usedmem;
		
		// If there is very little (or no) physical RAM left, then we'll
		// steal 1/16 of the holdable pages (e.g. 1M on a 16M machine)
		if (useable < holdable/16) useable = holdable/16;

		// If our initial freemem value is more than this amount
		// that we deem sensible to allocate, reduce it
		if (freemem > useable) freemem = useable;
		}
	return(freemem);
	}

/*	
	This routine is the first one called when an ODOC or PDOC event is received.
	
	In this routine you would place code used to setup structures, etc.
	which would be used in a 'for all docs' situation (like "Archive all
	dropped files")

	Obviously, the opening boolean tells you whether you should be opening
	or printing these files based on the type of event recieved.
	
	NEW IN 2.0!
	The itemCount parameter is simply the number of items that were dropped on
	the application and that you will be processing.  This gives you the ability
	to do a single preflight for memory allocation needs, rather than doing it
	once for each item as in previous versions.
	
	userDataHandle is a handle that you can create & use to store your own
	data structs.  This dataHandle will be passed around to the other
	odoc/pdoc routines so that you can get at your data without using
	globals - just like the new StandardFile.
	
	We also return a boolean to tell the caller if you support this type
	of event.  By default, our dropboxes don't support the pdoc, so when
	opening is FALSE, we return FALSE to let the caller send back the
	proper error code to the AEManager.

	You will probably want to remove the #pragma unused (currently there to fool the compiler!)
*/

#pragma segment Main
pascal Boolean PreFlightDocs (Boolean opening, long itemCount, Handle *userDataHandle) {
#pragma unused ( itemCount )
#pragma unused ( userDataHandle )
	OSErr resultCode1 = noErr, resultCode2 = noErr;
	
	ICError ICerr;
	ICAttr attr;
	Size maxTM, maxHM, growTM = 0, growHM = 0;
	int i,j;
	
	if (!opening) return(FALSE);	// we support opening, but not printing - see above
	
	MakeStatusWindow();
	ICerr = ICStart(&IC_instance, 'StUU');
#ifndef TARGET_API_MAC_CARBON
	if (!ICerr) ICerr = ICFindConfigFile(IC_instance, 0, nil);
#endif
	if (!ICerr) ICerr = ICGetPrefHandle(IC_instance, kICMapping, &attr, &IC_mappings);
	if (ICerr) { (void)Alert(Alert_NoIC, NULL); IC_mappings = NULL; }

	// See how much free memory we have in the temporary pool, and in the application heap
	maxTM = sensibleTempMaxMem(&growTM) & ~0xFFF;		// Round down to a multiple of 4K
	maxHM = MaxMem(&growHM)             & ~0xFFF;

	// Make sure we keep 32K of application memory in reserve for alerts etc.
	if (maxHM < 0x8000) maxHM = 0; else maxHM -= 0x8000;

	// Note: We lock all the handles as soon as possible, otherwise the System has
	// a habit of trying to copy them around to compact the heap, and that absolutely
	// kills performance. If we don't lock the handles, on a machine with VM the
	// launch time can be 10-20 seconds, even using the sensibleTempMaxMem routine.
	if (maxTM > maxHM)
		{
		InHandleSize = maxTM >> 1;
		OutHandleSize = InHandleSize >> 1;
		while (InHandleSize > 0x1000)
			{
			InBuffer  = TempNewHandle(InHandleSize, &resultCode1);
			if (resultCode1) { InHandleSize >>= 1; continue; }
			HLock(InBuffer);
			break;
			}
		while (OutHandleSize > 0x1000)
			{
			OutBuffer1 = TempNewHandle(OutHandleSize, &resultCode2);
			if (resultCode2) { OutHandleSize >>= 1; continue; }
			HLock(OutBuffer1);
			OutBuffer2 = TempNewHandle(OutHandleSize, &resultCode2);
			if (resultCode2) { DisposeHandle(OutBuffer1); OutBuffer1 = NULL; OutHandleSize >>= 1; continue; }
			HLock(OutBuffer2);
			break;
			}
		}
	else
		{
		InHandleSize = maxHM >> 1;
		OutHandleSize = InHandleSize >> 1;
		while (InHandleSize > 0x1000)
			{
			InBuffer  = NewHandle(InHandleSize);
			if (!InBuffer) { InHandleSize >>= 1; continue; }
			HLock(InBuffer);
			break;
			}
		while (OutHandleSize > 0x1000)
			{
			OutBuffer1 = NewHandle(OutHandleSize);
			if (!OutBuffer1) { InHandleSize >>= 1; continue; }
			HLock(OutBuffer1);
			OutBuffer2 = NewHandle(OutHandleSize);
			if (!OutBuffer2) { DisposeHandle(OutBuffer1); OutBuffer1 = NULL; OutHandleSize >>= 1; continue; }
			HLock(OutBuffer2);
			break;
			}
		resultCode1 = (InBuffer   == NULL);
		resultCode2 = (OutBuffer2 == NULL);
		}
	if (resultCode1 || resultCode2) { (void)Alert(Alert_NoMemory, NULL); StUU_Exit(); return(FALSE); }

	OutBuffer = OutBuffer1;
	ShowWindow(StatusWindow);
	PaintStatusWindow();
	
	j=0;
	for (i= 0; i<0x100; i++) B64decode_table[i] = 0xFF;
	for (i='A'; i<='Z'; i++) B64decode_table[i] = (unsigned char)j++;
	for (i='a'; i<='z'; i++) B64decode_table[i] = (unsigned char)j++;
	for (i='0'; i<='9'; i++) B64decode_table[i] = (unsigned char)j++;
	B64decode_table['+'] = (unsigned char)j++;
	B64decode_table['/'] = (unsigned char)j++;
	
#if DO_TIMING
	read_time  .lo = read_time  .hi = 0;
	create_time.lo = create_time.hi = 0;
	write_time .lo = write_time .hi = 0;
	close_time .lo = close_time .hi = 0;
	Microseconds(&total_start_time);
#endif
	return(TRUE);
}


/*	
	This routine is called for each file passed in the ODOC event.
	
	In this routine you would place code for processing each file/folder/disk that
	was dropped on top of you.
	
	You will probably want to remove the #pragma unused (currently there to fool the compiler!)
*/
#pragma segment Main
pascal void OpenDoc ( FSSpecPtr myFSSPtr, Boolean opening, Handle userDataHandle ) {
#pragma unused ( myFSSPtr )
#pragma unused ( opening )
#pragma unused ( userDataHandle )
	OSErr	err = noErr;
	
	
	#ifdef qWalkFolders
	/*
		For this case we need to determine if the FSSpec is a file or folder.
		If it's a folder, we then need to process each item in that folder,
		otherwise just process the item.
	*/
	if (FSpIsFolder(myFSSPtr))
		err = ProcessFolder(myFSSPtr);
	else
		err = ProcessItem(myFSSPtr);
	#else
	/*
		For this case we just call ProcessItem on the FSSpec above.
	*/
	err = ProcessItem(myFSSPtr);
	#endif
	
	// you should probably do something if you get back an error ;)
}

// CopyData Benchmarks
//  82 182 number of bytes (copied 100,000 times)
// --- ---
// 145 212 BlockMove
// 145 212 BlockMoveData
// 134 283 CopyData
// 113 237 CopyData in assembler
// Conclusion: For small blocks, a subroutine is quicker than the BlockMove trap

static void CopyData(register Ptr src, register Ptr dest, register unsigned short len)
	{
#if THINK_C
	asm	{
		sub.w	#1, len
	@0	move.b	(src)+, (dest)+
		dbra	len, @0
		}
#else
	while(len) { *dest++ = *src++; len--; }
#endif
	}

static void ShowWindowMsg(short y, unsigned char *msg, long time)
	{
	unsigned char buffer[128], number[16];
	NumToString(time, number);
	CopyData((Ptr)msg    + 1, (Ptr)buffer + 1, msg[0]);
	CopyData((Ptr)number + 1, (Ptr)buffer + 1 + msg[0], number[0]);
	buffer[0] = msg[0] + number[0];
	MoveTo(20, y); DrawString(buffer);
	}


/*	
	This routine is the last routine called as part of an ODOC event.
	
	In this routine you would place code to process any structures, etc.
	that you setup in the PreflightDocs routine.

	NEW IN 2.0!
	The itemCount parameter was the number of items that you processed.
	It is passed here just in case you need it ;)
	
	If you created a userDataHandle in the PreFlightDocs routines, this is
	the place to dispose of it since the Shell will NOT do it for you!
	
	You will probably want to remove the #pragma unusued (currently there to fool the compiler!)
*/

#pragma segment Main
pascal void PostFlightDocs ( Boolean opening, long itemCount, Handle userDataHandle ) {
#pragma unused ( opening )
#pragma unused ( itemCount )
#pragma unused ( userDataHandle )

	if ( (opening) && (!gOApped) )
		gDone = true;	//	close everything up!

#if DO_TIMING
	Microseconds(&total_end_time);
	WideSubtract((wide*)&total_end_time, (wide*)&total_start_time);
	EraseRect(&StatusWindow->portRect);
	MoveTo(20, 30); DrawString("\p(Times in milliseconds)");
	ShowWindowMsg( 50, "\pRead: ",   read_time.lo/1000L);
	ShowWindowMsg( 70, "\pCreate: ", create_time.lo/1000L);
	ShowWindowMsg( 90, "\pWrite: ",  write_time.lo/1000L);
	ShowWindowMsg(110, "\pClose: ",  close_time.lo/1000L);
	ShowWindowMsg(130, "\pTotal: ",  total_end_time.lo/1000L);
	while (1)
		{
		EventRecord e;
		WaitNextEvent(mDownMask, &e, 0, NULL);
		if (e.what == mouseDown) { FlushEvents(everyEvent,0); break; }
		}
#endif

	StUU_Exit();
	
	/*
		The reason we do not auto quit is based on a recommendation in the
		Apple event Registry which specifically states that you should NOT
		quit on a 'pdoc' as the Finder will send you a 'quit' when it is
		ready for you to do so.
	*/
}

// *********************************************************************************
// Main StUU Code body starts here

#define DebugStr(X)
#define debugstr(X)

// I treat any non-printing ascii character as a line break (non-printing ascii
// characters except for CR and/or NL have no business being in UUencoded data anyway)
#define IS_LINE_BREAK(C) ((unsigned char)(C) < (unsigned char)32)

// ReadNextLine silently eats any blank lines
// (saves having to distinguish between DOS/Mac/Unix line breaks).
static Ptr ReadNextLine(void)
	{
	register Ptr tempptr;
	register Ptr InPtrLocal = InPtr;			// Put a local copy of inPtr in a register
	register Ptr InPtrLimit = InPtrLocal + 256;
	
	// If data is all contiguous in memory use fast optimistic case
	if (InPtrLimit < InEnd)
		{
		if (IS_LINE_BREAK(*InPtrLocal)) InPtrLocal++;	// Skip initial line break, if present
		tempptr = InPtrLocal;							// Record start point
		
		// Search for first line break
		while (InPtrLocal < InPtrLimit && !IS_LINE_BREAK(*InPtrLocal)) InPtrLocal++;

		// If we have not hit the limit, and we read a non-blank line
		if (InPtrLocal < InPtrLimit && InPtrLocal > tempptr)
			{
			LineLength = InPtrLocal - tempptr;		// Set the length
			*InPtrLocal++ = 0;						// Zero-terminate the line
			InPtr = InPtrLocal;						// Update InPtr
			return(tempptr);						// and return
			}

		InPtrLocal = tempptr;	// Restore InPtrLocal before continuing with slow mechanism
		}

	// If the optimistic case fails, try the slow mechanism
	InPtrLimit = InEnd;
	tempptr = TempLineBuffer;

	while (1)
		{
		while (InPtrLocal < InPtrLimit && !IS_LINE_BREAK(*InPtrLocal) &&
			tempptr < &TempLineBuffer[255]) *tempptr++ = *InPtrLocal++;
		
		// We haven't run out of data yet, that means we broke out because we
		// found a line break, or we hit the line length limit.
		if (InPtrLocal < InPtrLimit)
			{
			if (IS_LINE_BREAK(*InPtrLocal)) InPtrLocal++;	// Skip over the newline
			// Test to see if the line is has some contents (empty lines are ignored)
			if (tempptr > TempLineBuffer) break;
			}
		else	// We broke out of loop because this chunk is over, so read some more
			{
			OSErr err;
			long inOutCount = BytesRemaining;
			if (inOutCount > InHandleSize) inOutCount = InHandleSize;
			if (inOutCount == 0) break;		// If no more to read, stop
#if DO_TIMING
			Microseconds(&read_start_time);
#endif
			UpdateStatusWindow(TRUE);
			err = FSRead(ReadRefNum, &inOutCount, *InBuffer);
			UpdateStatusWindow(FALSE);
#if DO_TIMING
			Microseconds(&read_end_time);
			WideSubtract((wide*)&read_end_time, (wide*)&read_start_time);
			WideAdd((wide*)&read_time, (wide*)&read_end_time);
#endif
			if (err) break;					// If error reading, stop
			InPtrLocal = *InBuffer;
			InPtrLimit = InEnd = InPtrLocal + inOutCount;
			BytesRemaining -= inOutCount;
			if (inOutCount == 0) { DebugStr("\pError: FSRead read nothing"); break; }
			}
		}

	LineLength = tempptr - TempLineBuffer;	// Set the length
	*tempptr = 0;							// Terminate the line
	InPtr = InPtrLocal;						// Update InPtr
	return(TempLineBuffer);
	}

// Returns true if the MIME types matchs exactly, OR
// if the candidate is exactly the major type of the target's major/minor pair.
// E.g. if the target is "audio/ulaw",
// then "audio/ulaw" and "audio" are the only two possible successful candidates.
static Boolean MatchMIME(const unsigned char *candidate, const unsigned char *target)
	{
	int i   = 1;
	int len = candidate[0];
	
	// If candidate MIME type is longer than the target MIME type, this cannot be a match
	if (len > target[0]) return(false);
	
	// Match characters until we run out, OR we find a difference, OR we find a '/'
	while (i <= len && candidate[i] == target[i] && candidate[i] != '/') i++;
	
	// If we matched the entire candidate string without finding a '/', this is a major-type match
	if (i > len) { DebugStr("\pMatchMIME: Major match;g"); DebugStr(target); }
	if (i > len) return(true);
	
	// Now we're looking for a full major/minor match, so the lengths must be identical
	if (len != target[0]) return(false);
	
	// Match characters until we run out, OR we find a difference
	while (i <= len && candidate[i] == target[i]) i++;
	
	// If we matched the entire candidate string, this is a full major/minor match
	if (i > len) return(true);
	
	// We failed to find a match
	return(false);
	}

#ifndef TARGET_API_MAC_CARBON
static ICError ICMapEntriesMIME(ICInstance inst, Handle entries, const unsigned char *mimetype, ICMapEntry *entry)
	{
	long posIndex = 0, posOfLongestLenSoFar = 0;
	unsigned char longestLenSoFar = 0;
	while (ICGetMapEntry(inst, entries, posIndex, entry) == noErr)
		{
		if (entry->MIME_type[0] > longestLenSoFar &&
			!(entry->flags & ICmap_not_incoming_mask) &&
			MatchMIME(entry->MIME_type, mimetype))
				{ longestLenSoFar = entry->MIME_type[0]; posOfLongestLenSoFar = posIndex; }
		posIndex += entry->total_length;
		}
	if (longestLenSoFar)
		return(ICGetMapEntry(inst, entries, posOfLongestLenSoFar, entry));
	return(icPrefNotFoundErr);
	}

static void GetFileTypeFromMIMEType(const unsigned char *mime, OSType *t, OSType *c)
	{
	ICMapEntry entry;
	ICError ICerr = -1;
	*t = *c = '????';
	if (IC_mappings) ICerr = ICMapEntriesMIME(IC_instance, IC_mappings, mime, &entry);
	if (ICerr == noErr) { *t = entry.file_type; *c = entry.file_creator; }

	if (ICerr == noErr)
		{
		DebugStr("\pContent-Type type/creator unspecified -- MIME match found;g");
		DebugStr(mime);
		}
	else
		{
		DebugStr("\pContent-Type type/creator unspecified -- no MIME match found;g");
		DebugStr(mime);
		}
	}

static void GetFileTypeFromFilename(const unsigned char *name, OSType *t, OSType *c)
	{
	ICMapEntry entry;
	ICError ICerr = -1;
	*t = *c = '????';
	if (IC_mappings) ICerr = ICMapEntriesFilename(IC_instance, IC_mappings, name, &entry);
	if (ICerr == noErr) { *t = entry.file_type; *c = entry.file_creator; }
	}
#endif

static pascal void CreateComplete(ParamBlockRec *p)
	{
	Create_HParamBlockRec *pb = (Create_HParamBlockRec*)p;
	if (pb->io.ioParam.ioResult == dupFNErr)			// Duplicate name; try again
		{
		OSErr err;
		pb->theSpec->name[pb->output_namesuffix_length  ] = pb->output_namesuffix++;
		pb->theSpec->name[pb->output_namesuffix_length-1] = '.';
		pb->theSpec->name[0] = pb->output_namesuffix_length;
		if (pb->output_namesuffix == '9'+1) pb->output_namesuffix = 'A';
		err = PBHCreateAsync(&pb->io);
		if (err) pb->io.ioParam.ioResult = err;
		}
	else pb->FileBeingCreated = FALSE;
	}

#if __MC68K__
static pascal void CreateCompleteRD(void)
	{
	Create_HParamBlockRec *pb = GetRegisterA0();
	void *oldGlobals = SetGlobalsRegister(pb->GlobalsReg);
	CreateComplete(pb);
	SetGlobalsRegister(oldGlobals);
	}
#else
//static RoutineDescriptor CreateCompleteRD =
//	BUILD_ROUTINE_DESCRIPTOR(uppIOCompletionProcInfo, CreateComplete);
//static IOCompletionUPP = NewIOCompletionUPP(CreateComplete);
#endif

// StUU_Create tries to create the file name indicated in TargetSpec
// Upon completion, createSpec will hold the spec that was actually created
// -- which may be different if the name had to be modified to make it unique.
// They have different uses:
// TargetSpec is the proposed Macintosh name
// fullname is the original file name (up to 256 characters) which should be used for things
// like the IC type lookup
// createSpec is the actual name created on disk, and should be used for things like setting
// the file info, opening the file for writing, and deleting the file to clean up.
// after a disk error

enum { StUU_Create_DataFork = 0, StUU_Create_ResourceFork, StUU_Overlay_DataFork };

static OSErr StUU_Create(const FSSpec * const TargetSpec, unsigned char *fullname, unsigned short mode)
	{
	OSErr err = noErr;
	unsigned long the_time;

#if DO_TIMING
	Microseconds(&create_start_time);
#endif

	// Only expect an AppleSingle Header if we're creating a resource fork
	// (i.e. decoding an application/applefile)
	NeedAppleSingleHeader = (mode == StUU_Create_ResourceFork);
	
	// If we're doing either fork of an AppleSingle decode then we need to set the file info
	// *after* we close the file, or else the last-modified time gets set to the time now.
	SetFileInfoAfterClose = (mode != StUU_Create_DataFork);

	// If we're creating a new resource fork, then remember the name, else clear it
	if (mode != StUU_Create_ResourceFork) B64_lastresourcefilename[0] = 0;

	// If we're creating a resource or data file, then we need to create the file
	// (and perhaps look up the file type) here. If we're doing a data fork overlay,
	// then we must have set the file type already (at the very least, we must already
	// know the file name, which wxill have determined the file type).
	if (mode != StUU_Overlay_DataFork)
		{
#ifndef TARGET_API_MAC_CARBON
		OSType ic_type, ic_creator;
#endif
		Create_HParamBlockRec pb;

		createSpec = *TargetSpec;		// Start with a copy of the targetspec

		// Fire off the asynchronous file creation
		pb.io.ioParam.ioCompletion = NewIOCompletionUPP(CreateComplete);//&CreateCompleteRD;
		pb.io.ioParam.ioResult     = noErr;
		pb.io.ioParam.ioNamePtr    = createSpec.name;
		pb.io.ioParam.ioVRefNum    = createSpec.vRefNum;
		pb.io.fileParam.ioDirID    = createSpec.parID;
		pb.theSpec = &createSpec;
		// Prepare the name suffix, in case we need it because of duplicate file names
		pb.output_namesuffix = '1';
		pb.output_namesuffix_length = createSpec.name[0]+2;
		if (pb.output_namesuffix_length > 31) pb.output_namesuffix_length = 31;
		pb.FileBeingCreated = TRUE;
		pb.GlobalsReg = GetGlobalsRegister();
		err = PBHCreateAsync(&pb.io);
		if (err) { ErrorAlert(kErrStringID, kCantCreateErr, err); return(err); }

		// Ask Internet Config to tell us the type and creator codes for this file
		// (This is done here so that the Create can run asynchronously while IC is thinking...)
		// We only have to do this if we don't already know the type and creator codes.
		// In the case of MIME files, we can get the type and creator from explicit
		// "x-mac-type" and "x-mac-creator" tags, or, failing that, we may have been
		// able to infer them from the MIME "type/subtype" information.
		if (!MacTypeCode || !CreatorCode)
			{
#ifndef TARGET_API_MAC_CARBON
			GetFileTypeFromFilename(fullname, &ic_type, &ic_creator);
			if (!MacTypeCode) MacTypeCode = ic_type;
			if (!CreatorCode) CreatorCode = ic_creator;
#endif
			}

		// Wait for file creation to complete.
		// We have to wait for the create to complete because until then we don't know
		// what name we're going to do the SetFInfo for.
		while (pb.FileBeingCreated) continue;

		if (pb.io.ioParam.ioResult)
			{
			ErrorAlert(kErrStringID, kCantCreateErr, pb.io.ioParam.ioResult);
			return(pb.io.ioParam.ioResult);
			}
		
		// Make sure previous PBHSetFInfoAsync is complete before we start smashing the parameter block
		while (Async_Info_PB.ioParam.ioResult == 1) continue;

		// Set the creator and type
		GetDateTime(&the_time);
		Async_Info_PB.fileParam.ioCompletion = NULL;
		Async_Info_PB.fileParam.ioNamePtr    = createSpec.name;
		Async_Info_PB.fileParam.ioVRefNum    = createSpec.vRefNum;
		Async_Info_PB.fileParam.ioFVersNum   = 0;
		Async_Info_PB.fileParam.ioFDirIndex  = 0;
		Async_Info_PB.fileParam.ioDirID      = createSpec.parID;
		Async_Info_PB.fileParam.ioFlCrDat    = the_time;
		Async_Info_PB.fileParam.ioFlMdDat    = the_time;
		Async_Info_PB.fileParam.ioFlFndrInfo.fdType     = MacTypeCode;
		Async_Info_PB.fileParam.ioFlFndrInfo.fdCreator  = CreatorCode;
		Async_Info_PB.fileParam.ioFlFndrInfo.fdFlags    = 0;
		Async_Info_PB.fileParam.ioFlFndrInfo.fdLocation = zeroPoint;
		Async_Info_PB.fileParam.ioFlFndrInfo.fdFldr     = 0;

		if (!SetFileInfoAfterClose) err = PBHSetFInfoAsync(&Async_Info_PB);
		}
	
	if (!err)
		{
		// Open the file
		Async_Open_PB.fileParam.ioCompletion = NULL;
		Async_Open_PB.fileParam.ioNamePtr    = createSpec.name;
		Async_Open_PB.fileParam.ioVRefNum    = createSpec.vRefNum;
		Async_Open_PB.fileParam.ioDirID      = createSpec.parID;
		Async_Open_PB.ioParam.ioPermssn      = fsWrPerm;
		if (mode == StUU_Create_ResourceFork) err = PBHOpenRFAsync(&Async_Open_PB);
		else                                  err = PBHOpenDFAsync(&Async_Open_PB);
		}

#if DO_TIMING
	Microseconds(&create_end_time);
	WideSubtract((wide*)&create_end_time, (wide*)&create_start_time);
	WideAdd((wide*)&create_time, (wide*)&create_end_time);
#endif

	// If the SetFInfo or the Open gave any (immediate) error, delete the file
	// and report the error
	if (err) { FSpDelete(&createSpec); ErrorAlert(kErrStringID, kCantCreateErr, err); }
	else UpdateStatusWindowOutput(createSpec.name);

	return(err);
	}

static void ReportCloseErrorAndDelete(void)
	{
	GrafPtr savePort;
	GetPort(&savePort);
	FSpDelete(&closeSpec);
	ParamText(closeSpec.name, NULL, NULL, NULL);
	(void)Alert(Alert_DiskError, NULL);
	SetPort(savePort);
	}

typedef union
	{
	ASIconBW     ic;		// entry ID 5
	ASFileDates  fd;		// entry ID 8
	ASFinderInfo fi;		// entry ID 9
	ASMacInfo    mi;		// entry ID 10
	} AppleSingleEntryData;

// Returns address of where resource fork data starts
static Ptr DecodeAppleSingleHeader(Ptr data)
	{
	uint16 i;
	//CInfoPBRec pb;
	AppleSingle *as = (AppleSingle *)data;
	Ptr resourcedata = data;

	DebugStr("\pWriting B64 resource fork");

	if (as->header.magicNum == 0x00051600) DebugStr("\pAppleSingle");
	else if (as->header.magicNum == 0x00051607) DebugStr("\pAppleDouble");
	else DebugStr("\pSomething else");
	
	if (as->header.versionNum != 0x00020000) DebugStr("\pWrong version");
	
	for (i=0; i<16; i++) if (as->header.filler[i]) DebugStr("\pFiller should be zero");

	//pb.hFileInfo.ioNamePtr    = createSpec.name;
	//pb.hFileInfo.ioVRefNum    = createSpec.vRefNum;
	//pb.hFileInfo.ioFDirIndex  = 0;
	//pb.hFileInfo.ioDirID      = createSpec.parID;
	//err = PBGetCatInfo(&pb, FALSE);

	for (i = 0; i < as->header.numEntries; i++)
		{
		AppleSingleEntryData *entrydata = (AppleSingleEntryData *)(data + as->entry[i].entryOffset);
		switch (as->entry[i].entryID)
			{
			case AS_DATA      : DebugStr("\pAS_DATA (handled)");
								if ((Ptr)entrydata <= OutPtr && (Ptr)entrydata + as->entry[i].entryLength <= OutPtr)
									{
									short fRefNum;
									OSErr err = FSpOpenDF(&writeSpec, fsWrPerm, &fRefNum);
									if (err == noErr)
										{
										long dummylength = as->entry[i].entryLength;
										err = FSWrite(fRefNum, &dummylength, (Ptr)entrydata);
										FSClose(fRefNum);
										}
									// If we got an error, signal it via the existing async error result
									if (err) Async_Write_PB.ioParam.ioResult = err;
									}
								else Async_Write_PB.ioParam.ioResult = -1;
								break;

			case AS_RESOURCE  : DebugStr("\pAS_RESOURCE (handled)");
								if ((Ptr)entrydata <= OutPtr) resourcedata = (Ptr)entrydata;
								else Async_Write_PB.ioParam.ioResult = -1;
								break;

			case AS_REALNAME  : {
								Str255 s;
								DebugStr("\pAS_REALNAME");
								s[0] = as->entry[i].entryLength;
								BlockMoveData((Ptr)entrydata, s+1, s[0]);
								DebugStr(s);
								}
								break;

			case AS_COMMENT   : {
								Str255 s;
								DebugStr("\pAS_COMMENT");
								s[0] = as->entry[i].entryLength;
								BlockMoveData((Ptr)entrydata, s+1, s[0]);
								if (s[0]) DebugStr(s);
								else DebugStr("\p<<Empty Comment String>>");
								}
								break;

			case AS_ICONBW    : DebugStr("\pAS_ICONBW"); break;			// Only provided for display purposes on non-Mac systems
			case AS_ICONCOLOR : DebugStr("\pAS_ICONCOLOR"); break;		// Only provided for display purposes on non-Mac systems
			case AS_FILEDATES : DebugStr("\pAS_FILEDATES (handled)");
								if (as->entry[i].entryLength >= sizeof(ASFileDates) && (Ptr)entrydata + sizeof(ASFileDates) <= OutPtr)
									{
									// Add 96 years plus 24 leap days to convert from AppleSingle's
									// 2000-based dates to Mac OS's 1904-based dates
									//pb.hFileInfo.ioFlCrDat = (96L*365L+24) *24L*60L*60L + entrydata->fd.create;
									//pb.hFileInfo.ioFlMdDat = (96L*365L+24) *24L*60L*60L + entrydata->fd.modify;
									Async_Info_PB.fileParam.ioFlCrDat = (96L*365L+24) *24L*60L*60L + entrydata->fd.create;
									Async_Info_PB.fileParam.ioFlMdDat = (96L*365L+24) *24L*60L*60L + entrydata->fd.modify;
									}
								break;
			case AS_FINDERINFO: DebugStr("\pAS_FINDERINFO (partially handled)");
								if (as->entry[i].entryLength >= sizeof(ASFinderInfo) && (Ptr)entrydata + sizeof(ASFinderInfo) <= OutPtr)
									{
									Async_Info_PB.fileParam.ioFlFndrInfo  = entrydata->fi.ioFlFndrInfo;
									//pb.hFileInfo.ioFlXFndrInfo = entrydata->fi.ioFlXFndrInfo;
									}
								break;
			case AS_MACINFO   : DebugStr("\pAS_MACINFO (ignored)");
								// Should, at most, probably only set the lock attribute here anyway
								//if (as->entry[i].entryLength >= sizeof(ASMacInfo) && (Ptr)entrydata + sizeof(ASMacInfo) <= OutPtr)
								//	pb.hFileInfo.ioFlAttrib = entrydata->mi.ioFlAttrib;
								break;
			case AS_PRODOSINFO: DebugStr("\pAS_PRODOSINFO (ignored)"); break;	// Only used on Apple PRODOS
			case AS_MSDOSINFO : DebugStr("\pAS_MSDOSINFO (ignored)"); break;	// Only used on Microsoft DOS
			case AS_AFPNAME   : DebugStr("\pAS_AFPNAME (ignored)"); break;		// Only used on AppleShare servers
			case AS_AFPINFO   : DebugStr("\pAS_AFPINFO (ignored)"); break;		// Only used on AppleShare servers
			case AS_AFPDIRID  : DebugStr("\pAS_AFPDIRID (ignored)"); break;		// Only used on AppleShare servers
			default           : DebugStr("\pUnknown entryID"); break;
			}
		}
	//pb.hFileInfo.ioNamePtr    = createSpec.name;
	//pb.hFileInfo.ioVRefNum    = createSpec.vRefNum;
	//pb.hFileInfo.ioFDirIndex  = 0;
	//pb.hFileInfo.ioDirID      = createSpec.parID;
	//if (!err) err = PBSetCatInfo(&pb,FALSE);
	
	return(resourcedata);
	}

// WriteChunk reports an error if:
// the open it depends on has failed, or the previous write (possibly to previous file) failed.
static OSErr WriteChunk(void)
	{
	Ptr datastart = *OutBuffer;
	OSErr err;
	// Make sure open is complete, and if it failed, report it now
	// Can't proceed until the open is complete, because we need the ioRefNum to write to
	while (Async_Open_PB.ioParam.ioResult == 1) continue;
	if (Async_Open_PB.ioParam.ioResult) return(Async_Open_PB.ioParam.ioResult);

	// Make sure previous write is complete, and if it failed, abort processing now
	while (Async_Write_PB.ioParam.ioResult == 1) continue;
	if (Async_Write_PB.ioParam.ioResult) return(Async_Write_PB.ioParam.ioResult);
	
#if DO_TIMING
	Microseconds(&write_start_time);
#endif

	writeSpec = createSpec;		// So we know which file to delete in case of a write error
	
	if (NeedAppleSingleHeader)		// If we're decoding an application/applefile
		{
		NeedAppleSingleHeader = false;
		datastart = DecodeAppleSingleHeader(*OutBuffer);
		if (Async_Write_PB.ioParam.ioResult) return(Async_Write_PB.ioParam.ioResult);
		}
	
	Async_Write_PB.ioParam.ioCompletion = NULL;
	Async_Write_PB.ioParam.ioResult     = noErr;
	Async_Write_PB.ioParam.ioRefNum     = Async_Open_PB.ioParam.ioRefNum;
	Async_Write_PB.ioParam.ioBuffer     = datastart;
	Async_Write_PB.ioParam.ioReqCount   = OutPtr - datastart;
	Async_Write_PB.ioParam.ioPosMode    = fsAtMark;
	Async_Write_PB.ioParam.ioPosOffset  = 0;
	err = PBWriteAsync(&Async_Write_PB);
	if (err) Async_Write_PB.ioParam.ioResult = err;
	
	if (OutBuffer == OutBuffer1) OutBuffer = OutBuffer2; else OutBuffer = OutBuffer1;
	OutPtr = *OutBuffer;
	OutEnd = *OutBuffer + OutHandleSize;

#if DO_TIMING
		Microseconds(&write_end_time);
		WideSubtract((wide*)&write_end_time, (wide*)&write_start_time);
		WideAdd((wide*)&write_time, (wide*)&write_end_time);
#endif

	return(noErr);
	}

static OSErr FinishDecode(OSErr err)
	{
	OSErr closeErr;
	
	// Make sure open is complete
	// Can't proceed until the open is complete, because we need the ioRefNum to close the file
	while (Async_Open_PB.ioParam.ioResult == 1) continue;
	
	// If file failed to be opened, give the error message
	if (Async_Open_PB.ioParam.ioResult)
		{
		FSpDelete(&createSpec);
		ErrorAlert(kErrStringID, kCantCreateErr, Async_Open_PB.ioParam.ioResult);
		return(Async_Open_PB.ioParam.ioResult);
		}
	// Okay, we now have an open file

	// Write remaining data out to disk
	if (!err && OutPtr > *OutBuffer) err = WriteChunk();
	// The file is written asynchronously, so the write proceeds while we continue
	// decoding the next file

	// And start the async close
#if DO_TIMING
	Microseconds(&close_start_time);
#endif
	// Make sure previous close is complete
	while (Async_Close_PB.ioParam.ioResult == 1) continue;
	if (Async_Close_PB.ioParam.ioResult) ReportCloseErrorAndDelete();
	// If the previous close failed, report it now

	closeSpec = createSpec;	// So we know which file to delete in case of an error on close
	Async_Close_PB.ioParam.ioCompletion = NULL;
	Async_Close_PB.ioParam.ioResult     = noErr;
	Async_Close_PB.ioParam.ioRefNum     = Async_Open_PB.ioParam.ioRefNum;
	closeErr = PBCloseAsync(&Async_Close_PB);
	if (closeErr) Async_Close_PB.ioParam.ioResult = closeErr;
	// The file is closed asynchronously, so the close proceeds while we continue
	// decoding the next file
	
	if (SetFileInfoAfterClose)
		{
		// Make sure previous PBHSetFInfoAsync is complete
		while (Async_Info_PB.ioParam.ioResult == 1) continue;
		DebugStr("\pDoing deferred setfinfo now");
		PBHSetFInfoAsync(&Async_Info_PB);	// Don't really care if this fails... What would I do about it?
		}

#if DO_TIMING
	Microseconds(&close_end_time);
	WideSubtract((wide*)&close_end_time, (wide*)&close_start_time);
	WideAdd((wide*)&close_time, (wide*)&close_end_time);
#endif
	
	// If we had a write error, no point continuing with further files
	if (Async_Write_PB.ioParam.ioResult != 1) return(Async_Write_PB.ioParam.ioResult);
	
	return(err);
	}

// CreateOutputFolder is passed the FSSpec of the source file
// It adds ' ' to the name and creates a folder with that name
// and fills in the new volume/directory in the targetspec
static OSErr CreateOutputFolder(FSSpec fs, FSSpecPtr targetspec)
	{
	OSErr err;
	long newdirid;
	
	// Folder names, like files, can be up to 31 characters long.
	// Cap length at 29 so we have room to add our ' ' on the end.
	if (fs.name[0] > 29) fs.name[0] = 29;
	fs.name[++fs.name[0]] = ' ';				// Add ' ' to the name
	fs.name[++fs.name[0]] = '';

	do
		{
		err = FSpDirCreate(&fs, smSystemScript, &newdirid);
		if (err != dupFNErr) break;				// If good, break out,
		fs.name[fs.name[0]]++;					// else try a new name
		} while (fs.name[fs.name[0]]);

	if (!err)
		{
		targetspec->vRefNum = fs.vRefNum;
		targetspec->parID   = newdirid;
		}
	
	return(err);
	}

static OSErr prepare_output_folder(const FSSpec * const sourcefile, FSSpec * const TargetFSSpec)
	{
	OSErr err;
	// If no name, then this is the first file, so don't create a subfolder yet
	// If parIDs are different, we're already in a subfolder, so don't need to do anything
	if (TargetFSSpec->name[0] == 0 || TargetFSSpec->parID != sourcefile->parID) return(noErr);
	// If we have already created a file (TargetFSSpec has a name) and we created
	// it in the source directory (parID == source parID), then is time to make
	// a subfolder and move the file into it. Also rename the file back to
	// its original desired name -- no need to worry about unique names in this
	// case because we know this is the first item going into the new folder.
	err = CreateOutputFolder(*sourcefile, TargetFSSpec);
	if (!err) err = CatMove(createSpec.vRefNum, createSpec.parID,
							createSpec.name, TargetFSSpec->parID, NULL);
	if (!err) err = HRename(TargetFSSpec->vRefNum, TargetFSSpec->parID,
							createSpec.name, TargetFSSpec->name);
	return(err);
	}

// UU decode stuff starts here

// Note, the decode table must be unsigned instead of signed to prevent the C compiler
// from trying to sign-extend the bytes every time it ORs them into a long
static unsigned char zero_char;
static unsigned char UUdecode_table[256];
#define	UU_VALUE(c) (UUdecode_table[(unsigned char)(c)])	/* single character decode */

static void reset_default_table(void)
	{
	int i;
	for (i= 0; i< 32; i++) UUdecode_table[i] = 0;
	for (i=32; i< 96; i++) UUdecode_table[i] = (unsigned char)(i - 32);
	for (i=96; i<256; i++) UUdecode_table[i] = 0;
	UUdecode_table['`'] = UUdecode_table[' '];
	UUdecode_table['~'] = UUdecode_table['^'];
	zero_char = ' ';
	
	B64_filename[0] = 0;		// When resetting after a successful decode, clear out the old B64 filename
	}

// On entry, CurrentLine is the "table" line
static void read_translation_table(void)
	{
	unsigned char val = 0;
	int i;
	for (i=0; i<256; i++) UUdecode_table[i] = 0;
	while (CurrentLine[0] && val < 64)
		{
		unsigned char *p;
		CurrentLine = ReadNextLine();
		p = (unsigned char *)CurrentLine;
		while (p[LineLength - 1] == ' ') p[--LineLength] = 0;
		if (val == 0) zero_char = *p;
		while (*p && val < 64) UUdecode_table[*p++] = val++;
		}
	}

#define LEGAL_MAC_FILENAME_CHARACTER(X) ((X) != 0 && (X) != ':')
#define MAX_MAC_FILENAME_LENGTH 31

static void getmacfilename(Ptr src, FSSpec *TargetFSSpec, unsigned char *pstring)
	{
	Ptr ptr;
	while (*src == ' ') src++;			// Skip leading spaces

	TargetFSSpec->name[0] = pstring[0] = 0;

	// Firstly fill in the actual disk filename we're going to use for the FSSpec
	ptr = src;
	while (*ptr && TargetFSSpec->name[0] < MAX_MAC_FILENAME_LENGTH)	// Read filename
		{
		unsigned char c = (unsigned char) *ptr++;
		if (!LEGAL_MAC_FILENAME_CHARACTER(c)) c = '-';
		TargetFSSpec->name[++TargetFSSpec->name[0]] = c;
		}
	// Strip trailing spaces and other garbage from file name
	while (TargetFSSpec->name[0] > 0 && TargetFSSpec->name[TargetFSSpec->name[0]] <= ' ')
		TargetFSSpec->name[0]--;
	
	// If no name specified, use the default "unspecified" name
	if (!TargetFSSpec->name[0]) GetIndString(TargetFSSpec->name, kMiscStringID, kUnspecFilename);

	// Secondly copy the full filename for extension mapping purposes
	ptr = src;
	while (*ptr && pstring[0] < 255) pstring[++pstring[0]] = *ptr++;
	while (pstring[0] > 0 && pstring[pstring[0]] <= ' ') pstring[0]--;
	if (!pstring[0]) GetIndString(pstring, kMiscStringID, kUnspecFilename);
	}

static Boolean stringmatch_cs(Ptr data, Ptr string)
	{
	while (*string) if (*data++ != *string++) return(FALSE);
	return(TRUE);
	}

#define lcase(X) (((X) >= 'A' && (X) <= 'Z') ? (X) - 'A' + 'a' : (X))

static Boolean stringmatch_ci(Ptr data, Ptr string)
	{
	while (*string)
		{
		char c1 = *data++;
		char c2 = *string++;
		if (lcase(c1) != lcase(c2)) return(FALSE);
		}
	return(TRUE);
	}

static Boolean samestring(char *s1, char *s2)
	{
	while (*s1 || *s2) if (*s1++ != *s2++) return(FALSE);
	return(TRUE);
	}

static void copystring(char *s, char *d)
	{
	do *d++ = *s; while (*s++);
	}

/* Old Think C inline assembly chunk
#if THINK_C
		asm {
			clr.l	d0
			move.b	(p)+, d0			; read first byte
			move.b	(decoder,d0), val	; move translated value into val
			move.b	(p)+, d0			; read second byte
			lsl.l	#6, val				; shift val left six bits
			or.b	(decoder,d0), val	; OR translated value into val
			move.b	(p)+, d0			; read third byte
			lsl.l	#6, val				; shift val left six bits
			or.b	(decoder,d0), val	; OR translated value into val
			move.b	(p)+, d0			; read final byte
			lsl.l	#6, val				; shift val left six bits
			or.b	(decoder,d0), val	; OR translated value into val
			}
*/

static OSErr DecodeUULine(register Ptr p)
	{
	register Ptr out;
	register unsigned char *decoder = UUdecode_table;
	register unsigned long x;
	register int outputlen = decoder[(unsigned char)(*p++)];

	// UU text line can never encode more than 63 data bytes
	// The +2 is because our high-speed loop may overshoot by up to two
	if (OutEnd - OutPtr < outputlen+2) { OSErr err = WriteChunk(); if (err) return(err); }
	out = OutPtr;		// Careful! Don't set this till *after* WriteChunk

	while (outputlen > 0)
		{
		x  = decoder[(unsigned char)(p[0])]; x <<= 6;
		x |= decoder[(unsigned char)(p[1])]; x <<= 6;
		x |= decoder[(unsigned char)(p[2])]; x <<= 6;
		x |= decoder[(unsigned char)(p[3])];
		out[2] = (char)x; x >>= 8;
		out[1] = (char)x; x >>= 8;
		out[0] = (char)x;
		p += 4;
		out += 3;
		outputlen -= 3;
		}

	if (out > OutEnd) DebugStr("\pERROR! OutPtr overrun");
	if (out == OutEnd) DebugStr("\pNOTE: OutPtr exactly correct!");

	// Because we output blocks of three, we may overshoot the correct end of line
	// In this case outputlen will be negative and we backtrack OutPtr by that amount.
	OutPtr = out + outputlen;
	return(noErr);
	}

#define MAX_STASHED_LINES 8
#define MAX_STASHED_LINE_LENGTH 88
typedef struct { char c[MAX_STASHED_LINE_LENGTH]; } TextLine;

#define PERFECT_LINE (CurrentLine[0] == 'M' && LineLength >= 61 && LineLength <= 63)

// TargetFSSpec gives the target volume/folder to put the files into --
// DecodeUUFileData fills in the name part itself from the name given in the source file
static OSErr DecodeUUFileData(FSSpecPtr TargetFSSpec)
	{
	OSErr err;
	TextLine LastPerfectLine;
	TextLine LineStash[MAX_STASHED_LINES];
	int num_stashed_lines = 0;
	Boolean GotPerfectLine = FALSE;
	int i;
	Ptr filemode;
	Ptr hp = CurrentLine + 6;			// Skip over the "begin "
	unsigned char full_filename[256];

	while (*hp == ' ') hp++;			// Skip leading spaces
	filemode = hp;						// The Unix file permissions
	while (*hp && *hp != ' ') hp++;		// Read the permission portion
	if (*hp) *hp++ = 0;					// Terminate the permission string
	getmacfilename(hp, TargetFSSpec, full_filename);

	// Okay, we have consumed the header line now. Move on to the first line of encoded data
	CurrentLine = ReadNextLine();
	
	// Create output file
	// If the target filename already exists, then StUU_Create will
	// fabricate a new unique filename
	err = StUU_Create(TargetFSSpec, full_filename, StUU_Create_DataFork);
	if (err) return(err);
	// The file is opened asynchronously, so the open proceeds while we start decoding
	
	OutPtr = *OutBuffer;				// Set up output pointers
	OutEnd = *OutBuffer + OutHandleSize;
	
	while (CurrentLine[0] && !err)
		{
		// "Good Loop": Read as many good lines as possible
		while (PERFECT_LINE && !err)
			{ err = DecodeUULine(CurrentLine); CurrentLine = ReadNextLine(); }
		
		GotPerfectLine = FALSE;
		// "Bad Loop": Skip past the bad or questionable lines
		while (CurrentLine[0] && !err)
			{
			int origlen = UU_VALUE(CurrentLine[0]);			// Original length
			int codedlength = 1 + (origlen * 4 + 2) / 3;	// UUencoded length
			
			if (CurrentLine[0] == 'e' &&		// check for "end" marker
				CurrentLine[1] == 'n' &&
				CurrentLine[2] == 'd') goto end_of_file;
			
			if (CurrentLine[0] == 'b' &&		// check for another "begin" marker
				CurrentLine[1] == 'e' &&
				CurrentLine[2] == 'g' &&
				CurrentLine[3] == 'i' &&
				CurrentLine[4] == 'n' &&
				CurrentLine[5] == ' ') goto end_of_file;
			
			if (CurrentLine[0] == 't' &&		// check for another "table" marker
				CurrentLine[1] == 'a' &&
				CurrentLine[2] == 'b' &&
				CurrentLine[3] == 'l' &&
				CurrentLine[4] == 'e') goto end_of_file;

			// If line too short, restore truncated trailing spaces
			if (LineLength < codedlength)
				{
				// Mustn't try to expand line in place!
				if (CurrentLine != TempLineBuffer)
					{
					CopyData(CurrentLine, TempLineBuffer, (unsigned short)LineLength);
					CurrentLine = TempLineBuffer;
					}
				while (LineLength < codedlength) CurrentLine[LineLength++] = (char)zero_char;
				CurrentLine[LineLength] = 0;		// Terminate the line
				}
			
			// If we get one perfect line stash it, and set the flag.
			// If we get a second perfect line, assume we are back in body text,
			// and go back to the good loop
			if (PERFECT_LINE)
				{
				if (GotPerfectLine) { err = DecodeUULine(LastPerfectLine.c); break; }
				num_stashed_lines = 0;
				GotPerfectLine = TRUE;
				CopyData(CurrentLine, LastPerfectLine.c, (unsigned short)LineLength+1);
				// The +1 is because we need to copy the trailing null as well
				}
			else
				{
				// Stash runt lines in case they turn out to be
				// the last few lines before the "end" marker
				GotPerfectLine = FALSE;
				if (LineLength <= MAX_STASHED_LINE_LENGTH &&
					num_stashed_lines < MAX_STASHED_LINES &&
					origlen > 0 && origlen <= 45 && LineLength <= codedlength + 3)
						CopyData(CurrentLine, LineStash[num_stashed_lines++].c, (unsigned short)LineLength+1);
						// The +1 is because we need to copy the trailing null as well
				}
			CurrentLine = ReadNextLine();						// Read the next line
			}
		}

end_of_file:
	// Decode any stashed lines
	if (!err && GotPerfectLine) err = DecodeUULine(LastPerfectLine.c);
	for (i=0; !err && i<num_stashed_lines; i++) err = DecodeUULine(LineStash[i].c);

	return(FinishDecode(err));
	}

// Base64 stuff starts here

#define HEX_VALUE(X)  ( (X) >= '0' && (X) <= '9' ? (X)-'0'    :   \
						(X) >= 'A' && (X) <= 'F' ? (X)-'A'+10 :   \
						(X) >= 'a' && (X) <= 'f' ? (X)-'a'+10 : -1)

static OSType parsehexostype(char *p)
	{
	union { OSType ostype; unsigned char byte[4]; } result;
	result.byte[0] = HEX_VALUE(p[0]) << 4 | HEX_VALUE(p[1]);
	result.byte[1] = HEX_VALUE(p[2]) << 4 | HEX_VALUE(p[3]);
	result.byte[2] = HEX_VALUE(p[4]) << 4 | HEX_VALUE(p[5]);
	result.byte[3] = HEX_VALUE(p[6]) << 4 | HEX_VALUE(p[7]);
	return(result.ostype);
	}

// On entry, CurrentLine is the "Content-" line
// Returns true if it finds a "Content-Transfer-Encoding: base64" line
static Boolean read_b64_content_line(void)
	{
	if (stringmatch_ci(CurrentLine+8, "Type:"))
		{
		Ptr p = CurrentLine+13;
		B64_MIMEtype[0] = 0;
		//DebugStr("\pContent-Type");
		while (*p == ' ') p++;				// Skip spaces
		while (*p && *p != ' ' && *p != ';' && B64_MIMEtype[0] < sizeof(B64_MIMEtype)-2) // Leave space for trailing NULL
			B64_MIMEtype[++B64_MIMEtype[0]] = *p++;		// Copy MIME type string
		//DebugStr(B64_MIMEtype);
		if (*p == ';') p++;					// Skip ';'
		
		// Clear any previous type and creator codes
		// Don't clear B64_filename -- a "Content-Type: application/applefile" section relies
		// on the filename given in the previous "Content-Type: multipart/appledouble" header.
		MacTypeCode = CreatorCode = 0;
		
		while (*p)							// Process the keywords
			{
			char key[256], value[256], *dst;
			dst = key;
			while (*p == ' ') p++;					// Skip spaces
			while (*p && *p != '=') *dst++ = *p++;	// Copy the key
			*dst++ = 0;								// Null-terminate
			if (!*p) break;							// If no more data, bail out
			p++;									// Skip the '='
			if (*p != '"') break;					// If not a quote, bail out
			p++;									// Skip the quote
			dst = value;
			while (*p && *p != ';') *dst++ = *p++;	// Copy the value
			if (*p == ';') p++;						// Skip the ';'
			if (dst == value) break;
			dst--;
			if (*dst != '"') break;					// If last character not a quote, bail out
			*dst++ = 0;								// Null-terminate

			//debugstr(key);
			//debugstr(value);
			if      (samestring(key, "name"         )) CopyData(value, B64_filename, dst-value);
			else if (samestring(key, "x-mac-type"   )) MacTypeCode = parsehexostype(value);
			else if (samestring(key, "x-mac-creator")) CreatorCode = parsehexostype(value);
			}
		}
	else if (stringmatch_ci(CurrentLine+8, "Transfer-Encoding:"))
		{
		Ptr p = CurrentLine+26;
		//DebugStr("\pContent-Transfer-Encoding");
		while (*p == ' ') p++;
		if (stringmatch_ci(p, "base64")) return(TRUE);
		}
	return(FALSE);
	}

#define BadLine(X) return(FALSE);
//#define BadLine(X) do { DebugStr(X); return(FALSE); } while(0)

static Boolean valid_b64_line(void)
	{
	Ptr ptr = CurrentLine;
	Ptr end = CurrentLine + LineLength;
	long x;
	while (end > ptr && end[-1] == ' ') end--;	// Strip trailing spaces
	x = end - ptr; // Line length must be multiple of four, more than zero but not more than 76
	if (x &  3) BadLine("\pLine length not multiple of four");
	if (x == 0) BadLine("\pLine empty");
	if (x > 76) BadLine("\pLine length greater than 76");
	while (ptr < end && B64decode_table[*ptr] != 0xFF) ptr++;	// Check characters are valid
	if (ptr < end && *ptr == '=') ptr++;		// Allow up to two trailing '=' characters
	if (ptr < end && *ptr == '=') ptr++;
	if (ptr < end) BadLine("\pInvalid Base 64 character");
	return(TRUE);								// If we got to the end, it's valid
	}

static Boolean DecodeB64Line(void)
	{
	register Ptr p = CurrentLine;
	register Ptr out;
	register unsigned char *decoder = B64decode_table;
	register unsigned long x;
	register char y;
	register int outputlen;
	Ptr end = p + LineLength;
	
	while (end > p && end[-1] == ' ') end--;	// Strip trailing spaces
	x = end - p; // Line length must be multiple of four, more than zero but not more than 76
	if ((x & 3) || x == 0 || x > 76) return(FALSE);
	outputlen = (x>>2) * 3;
	if (end[-1] == '=')							// Allow up to two trailing '=' characters
		{
		end[-1] = 'A';
		outputlen--;
		if (end[-2] == '=') { end[-2] = 'A'; outputlen--; }
		}

	// Base64 line can never encode more than 76/4*3 = 57 data bytes
	// The +2 is because our high-speed loop may overshoot by up to two
	// (which we correct at the end)
	if (OutEnd - OutPtr < outputlen+2) { if (WriteChunk()) return(FALSE); }
	out = OutPtr;		// Careful! Don't set this till *after* WriteChunk

	while (outputlen > 0)
		{
		y = decoder[(unsigned char)(p[0])]; if (y<0) goto failure; x = ((unsigned long)y) << 6;
		y = decoder[(unsigned char)(p[1])]; if (y<0) goto failure; x |= y; x <<= 6;
		y = decoder[(unsigned char)(p[2])]; if (y<0) goto failure; x |= y; x <<= 6;
		y = decoder[(unsigned char)(p[3])]; if (y<0) goto failure; x |= y;
		out[2] = (char)x; x >>= 8;
		out[1] = (char)x; x >>= 8;
		out[0] = (char)x;
		p += 4;
		out += 3;
		outputlen -= 3;
		}

	if (out > OutEnd) DebugStr("\pERROR! OutPtr overrun");
	if (out == OutEnd) DebugStr("\pNOTE: OutPtr exactly correct!");

	// Because we output blocks of three, we may overshoot the correct end of line
	// In this case outputlen will be negative and we backtrack OutPtr by that amount.
	OutPtr = out + outputlen;
	return(TRUE);

failure:
	return(FALSE);
	}

// TargetFSSpec gives the target volume/folder to put the files into --
// DecodeB64FileData fills in the name part itself from the name given in the source file
static OSErr DecodeB64FileData(const FSSpec * const sourcefile, FSSpecPtr TargetFSSpec)
	{
	OSErr err;
	unsigned short mode;
	unsigned char full_filename[256];
	
	B64_MIMEtype[B64_MIMEtype[0]+1] = 0;
	if (samestring((char*)B64_MIMEtype+1, "application/applefile"))
		{
		// If we're creating a resource fork, then that assumes we're going to be
		// getting an AppleDouble header giving the file type and creator codes
		debugstr("StUU_Create_ResourceFork");
		mode = StUU_Create_ResourceFork;
		copystring(B64_filename, B64_lastresourcefilename);
		}
	else if (B64_filename[0] && samestring(B64_lastresourcefilename, B64_filename))
		{
		// If we're overlaying a data fork, then that implies that we already
		// got an AppleDouble header giving the file type and creator codes
		debugstr("StUU_Overlay_DataFork");
		mode = StUU_Overlay_DataFork;
		}
	else
		{
		// If we're creating a raw data fork from scratch, and we didn't get
		// our type and creator explicitly in the Content-Type line (priority 2),
		// then try to infer them from the MIME type instead (priority 3).
		debugstr("StUU_Create_DataFork");
		mode = StUU_Create_DataFork;
		if (!MacTypeCode || !CreatorCode)
			{
#ifndef TARGET_API_MAC_CARBON
			OSType ic_type = 0, ic_creator;
			GetFileTypeFromMIMEType(B64_MIMEtype, &ic_type, &ic_creator);
			if (!MacTypeCode) MacTypeCode = ic_type;
			if (!CreatorCode) CreatorCode = ic_creator;
#endif
			}
		}

	if (mode != StUU_Overlay_DataFork)
		{
		// Must do prepare_output_folder before we fill in the new TargetFSSpec filename
		err = prepare_output_folder(sourcefile, TargetFSSpec);
		if (err) return(err);
		getmacfilename(B64_filename, TargetFSSpec, full_filename);
		}

	DebugStr("\pDecodeB64FileData creating file;g");
	if (B64_filename[0]) debugstr(B64_filename);
	else debugstr("<<no name specified>>");

	// Create output file
	// If the target filename already exists, then StUU_Create will
	// fabricate a new unique filename
	err = StUU_Create(TargetFSSpec, full_filename, mode);
	if (err) return(err);
	// The file is opened asynchronously, so the open proceeds while we start decoding
	
	OutPtr = *OutBuffer;				// Set up output pointers
	OutEnd = *OutBuffer + OutHandleSize;
	
	while (CurrentLine[0] && DecodeB64Line()) CurrentLine = ReadNextLine();

	return(FinishDecode(err));
	}

// *********************************************************************************

/*
	This routine gets called for each item (which could be either a file or a folder)
	that the caller wants dropped.  The determining factor is the definition of the
	qWalkFolder compiler directive.   Either way, the item in question should be
	processed as a single item and not "dissected" into component units (like subfiles
	of a folder!)
*/
static OSErr ProcessItem(FSSpecPtr myFSSPtr)
	{
	OSErr err;
	FSSpec TargetFSSpec;
	Boolean B64_flag = FALSE;

	err = FSpOpenDF(myFSSPtr, fsRdPerm, &ReadRefNum);
	if (err) return(err);
	
	err = GetEOF(ReadRefNum, &BytesRemaining);
	if (err) { FSClose(ReadRefNum); return(err); }
	InPtr = InEnd = NULL;
	
	UpdateStatusWindowInput(myFSSPtr->name);
	
	TargetFSSpec.vRefNum = myFSSPtr->vRefNum;
	TargetFSSpec.parID   = myFSSPtr->parID;
	TargetFSSpec.name[0] = 0;
	
	reset_default_table();
	CurrentLine = ReadNextLine();		// Load first line of file into CurrentLine
	while (err == noErr && CurrentLine[0])
		{
		if (CurrentLine[0] == 'e' &&		// check for "end" marker
			CurrentLine[1] == 'n' &&
			CurrentLine[2] == 'd') { reset_default_table(); B64_flag = FALSE; }
		
		if (CurrentLine[0] == 't' &&
			CurrentLine[1] == 'a' &&
			CurrentLine[2] == 'b' &&
			CurrentLine[3] == 'l' &&
			CurrentLine[4] == 'e') { read_translation_table(); B64_flag = FALSE; }
			
		if (CurrentLine[0] == 'b' &&
			CurrentLine[1] == 'e' &&
			CurrentLine[2] == 'g' &&
			CurrentLine[3] == 'i' &&
			CurrentLine[4] == 'n' &&
			CurrentLine[5] == ' ')
				{
				B64_MIMEtype[0] = MacTypeCode = CreatorCode = 0;
				err = prepare_output_folder(myFSSPtr, &TargetFSSpec);
				if (!err) err = DecodeUUFileData(&TargetFSSpec);
				reset_default_table();
				B64_flag = FALSE;
				}
		
		if (CurrentLine[0] == 'C' &&
			CurrentLine[1] == 'o' &&
			CurrentLine[2] == 'n' &&
			CurrentLine[3] == 't' &&
			CurrentLine[4] == 'e' &&
			CurrentLine[5] == 'n' &&
			CurrentLine[6] == 't' &&
			CurrentLine[7] == '-') if (read_b64_content_line()) B64_flag = TRUE;
		
		if (B64_flag && valid_b64_line())
			{
			err = DecodeB64FileData(myFSSPtr, &TargetFSSpec);
			reset_default_table();
			B64_flag = FALSE;
			}
		
		CurrentLine = ReadNextLine();	// Read next line
		}
	FSClose(ReadRefNum);
	while (Async_Close_PB.ioParam.ioResult == 1) continue;	// Make sure final output close is complete
	while (Async_Info_PB.ioParam.ioResult == 1) continue;	// And the PBHSetFInfoAsync

	// If we exited because of a write error, report it and delete the file
	// (File will already have been closed for us by DecodeUUFileData routine)
	if (Async_Write_PB.ioParam.ioResult)
		{
		ErrorAlert(kErrStringID, kDiskFullErr, Async_Write_PB.ioParam.ioResult);
		FSpDelete(&writeSpec);
		ParamText(writeSpec.name, NULL, NULL, NULL);
		(void)Alert(Alert_DiskError, NULL);
		}
	// Report any close errors (should be pretty rare, I think).
	// (Don't bother reporting close error if we've already reported write error for this file)
	else if (Async_Close_PB.ioParam.ioResult) ReportCloseErrorAndDelete();
	
	// if we only created a single file, move it up a level and delete the folder
	
	return(err);
}

/*
	This routine gets called for any folder (or disk) that the caller wants
	processed as a set of component items, instead of as a single entity.
	The determining factor is the definition of the qWalkFolder compiler directive.
*/
static OSErr ProcessFolder(FSSpecPtr myFSSPtr)
{
	OSErr		err = noErr;
	short		index, oldIndex, localIndex;
	FSSpec		localFSSpec, curFSSpec;
	CInfoPBRec	cipb;
	Str255		fName, vFName;
	long		dirID, origDirID;
	Boolean		foundPosition;

 	// copy the source locally to avoid recursion problems
 	BlockMoveData(myFSSPtr, &localFSSpec, sizeof(FSSpec));
 	
	//	get the dirID for THIS folder, not it's parent!
	BlockMoveData(localFSSpec.name, fName, 32);
	
	cipb.hFileInfo.ioCompletion	= 0L;
	cipb.hFileInfo.ioNamePtr	= fName;
	cipb.hFileInfo.ioVRefNum	= localFSSpec.vRefNum;
	cipb.hFileInfo.ioFDirIndex	= 0;	// use the dir & vRefNum;
	cipb.hFileInfo.ioDirID		= localFSSpec.parID;
	err = PBGetCatInfoSync(&cipb);
	
	if (!err) {		
		origDirID = cipb.dirInfo.ioDrDirID; // copy the sucker
		index = 1;
				
		// index through all contents of this folder
		while (err == noErr) {
			dirID = origDirID;
			localIndex = index;
			fName [0] = 0;
			cipb.hFileInfo.ioCompletion	= 0L;
			cipb.hFileInfo.ioNamePtr	= fName;
			cipb.hFileInfo.ioVRefNum	= localFSSpec.vRefNum;
			cipb.hFileInfo.ioFDirIndex	= localIndex;	// use a real index
			cipb.hFileInfo.ioDirID		= dirID;
			err = PBGetCatInfoSync(&cipb);

			if (!err) {
				BlockMoveData(fName, curFSSpec.name, 32);
				curFSSpec.vRefNum	= cipb.hFileInfo.ioVRefNum;
				curFSSpec.parID		= dirID;
			
				/*	
					Check to see if this entry is a folder.
				*/
				if (cipb.hFileInfo.ioFlAttrib & ioDirMask) {
					err = ProcessFolder(&curFSSpec);
				 } else
					err = ProcessItem(&curFSSpec);
			
				/*	If we've had an error, get out! */
				if (err)	break;

				// dirID = origDirID;	
				localIndex = index;	

				/*	
					Now take into account new files being created
					in the current directory & messing up our index.
					See Dev.CD Vol. XI:Tools & Apps (Moof!):Misc Utilities:
					Disinfectant & Source 2.5.1:Sample:Notes:Scan Alg	
				*/
				vFName [0] = 0;
				cipb.hFileInfo.ioCompletion	= 0L;
				cipb.hFileInfo.ioNamePtr	= vFName;
				cipb.hFileInfo.ioVRefNum	= localFSSpec.vRefNum;
				cipb.hFileInfo.ioFDirIndex	= localIndex;	// use a real index
				cipb.hFileInfo.ioDirID		= dirID;
				err = PBGetCatInfoSync(&cipb);
				oldIndex = index;
				if (!err) {
					/*	If they're equal - same place, go to next */
					if (EqualString (vFName, fName, false, false))
						index++;
				}
				
				/*	If we didn't advance, then perhaps a file was created or deleted */
				if (oldIndex == index) {
					oldIndex		= index;	/* save off the old */
					index			= 0;		/* and start at the beginning */
					err				= noErr;
					vFName [0]		= 0;
					foundPosition	= false;
					
					while (!foundPosition) {
						index++;
						vFName [0] = 0;
						cipb.hFileInfo.ioCompletion	= 0L;
						cipb.hFileInfo.ioNamePtr	= vFName;
						cipb.hFileInfo.ioVRefNum	= localFSSpec.vRefNum;
						cipb.hFileInfo.ioFDirIndex	= index;	/* now use a real index */
						cipb.hFileInfo.ioDirID		= dirID;
						err = PBGetCatInfoSync(&cipb);
						
						if (err == fnfErr) {  // we've just been deleted
							index = oldIndex;
							foundPosition = true;
							err = noErr;	// have to remember to reset this!
						}
						
					/*	found same file & same index position */
					/*	so try the next item */
						if ((!foundPosition) && EqualString(fName, vFName, false, false)) {
							index++;
							foundPosition = true;
						}
					}
				}
			}
		}
	}
	
	return(err);
}

/*
	This routine is called when the user chooses "Select File" from the
	File Menu.
	
	Currently it simply calls the new StandardGetFile routine to have the
	user select a single file (any type, numTypes = -1) and then calls the
	SendODOCToSelf routine in order to process it.
			
	The reason we send an odoc to ourselves is two fold: 1) it keeps the code
	cleaner as all file openings go through the same process, and 2) if events
	are ever recordable, the right things happen (this is called Factoring!)

	Modification of this routine to only select certain types of files, selection
	of multiple files, and/or handling of folder & disk selection is left
	as an exercise to the reader.
*/
pascal void SelectFile(void)
{
#ifndef TARGET_API_MAC_CARBON
	StandardFileReply	stdReply;
	SFTypeList			theTypeList;

	StandardGetFile(NULL, -1, theTypeList, &stdReply);
	if (stdReply.sfGood)	// user did not cancel
		SendODOCToSelf(&stdReply.sfFile);	// so send me an event!
#endif
}

/*
	This routine is called during the program's initialization and gives you
	a chance to allocate or initialize any of your own globals that your
	dropbox needs.
	
	You return a boolean value which determines if you were successful.
	Returning false will cause DropShell to exit immediately.
*/
pascal Boolean InitUserGlobals(void)
{
	return(true);	// nothing to do, it we must be successful!
}

/*
	This routine is called during the program's cleanup and gives you
	a chance to deallocate any of your own globals that you allocated
	in the above routine.
*/
pascal void DisposeUserGlobals(void)
{
	// nothing to do for our sample dropbox
}
